#!/usr/bin/env python

# GhostScript interface API through ctypes
# The GhostScript Windows DLL is interfaced through gsdll32.dll
# Linux is libgs.so and OS X *should* be Ghostscript.framework
import os, sys
from ctypes import *
from gserrors import *

# Try to get the right library to load
# based on the OS
if sys.platform.startswith("win"):
    gs = windll.LoadLibrary("gsdll32.dll")
elif sys.platform.startswith("darwin"):
    gs = cdll.LoadLibrary("Ghostscript.framework")
elif sys.platform.startswith("linux"):
    gs = cdll.LoadLibrary("libgs.so")

class GSAPIVERSION(Structure):
    _fields_ = [
        ("product", c_char_p),
        ("copyright", c_char_p),
        ("revision", c_long),
        ("revisiondate", c_long )
    ]
# end class GSVersion

gs_version = gs.gsapi_revision
gs_version.argtypes = [POINTER(GSAPIVERSION), c_int]
gs_version.restype = c_int

def GhostscriptVersion():
    # Returns the version of Ghostscript currently in use
    _gs_version = GSAPIVERSION()
    gs_version( byref(_gs_version), c_int(sizeof(GSAPIVERSION())))
    return _gs_version
# end GhostscriptVersion

GSVersion = GhostscriptVersion()

# gsapi_new_instance -- Create a new instance handle
# Only 1 instance handle can be created at a time!
_gs_new_instance = gs.gsapi_new_instance
_gs_new_instance.argtypes = [POINTER(c_void_p), c_void_p]
_gs_new_instance.restype = c_int

# gsapi_init_with_args (void *instance, int argc, char **argv)
_gs_init = gs.gsapi_init_with_args
#_gs_init.argtypes = [POINTER(c_void_p), c_int, c_char_p)]
_gs_init.restype = c_int

# gsapi_exit(void *inst) -- Clear exit from GS
_gs_exit = gs.gsapi_exit
_gs_exit.argtypes = [c_void_p]
_gs_exit.restype = c_int

# gsapi_delete_instance -- Delete an instance handle to Ghostscript
_gs_delete_instance = gs.gsapi_delete_instance
_gs_delete_instance.argtypes = [c_void_p]

# The gsapi_run_* functions are like gs_main_run_* except that the error_object is omitted. If these functions 
# return <= -100, either quit or a fatal error has occured. You must call gsapi_exit() next. The only exception is 
# gsapi_run_string_continue() which will return e_NeedInput if all is well.

# The address passed in pexit_code will be used to return the exit code for the interpreter in case of a quit or 
# fatal error. The user_errors argument is normally set to zero to indicate that errors should be handled through 
# the normal mechanisms within the interpreted code. If set to a negative value, the functions will return an 
# error code directly to the caller, bypassing the interpreted language. The interpreted language's error handler 
# is bypassed, regardless of user_errors parameter, for the e_interrupt generated when the polling callback 
# returns a negative value. A positive user_errors is treated the same as zero.

# There is a 64 KB length limit on any buffer submitted to a gsapi_run_* function for processing. If you have more 
# than 65535 bytes of input then you must split it into smaller pieces and submit each in a separate 
# gsapi_run_string_continue() call. 

# int gsapi_run_string_begin (void *instance, int user_errors, int *pexit_code)
_gs_run_string_begin = gs.gsapi_run_string_begin
_gs_run_string_begin.argtypes = [c_void_p, c_int, POINTER(c_int)]
_gs_run_string_begin.restype = c_int

# int gsapi_run_string_continue (void *instance, const char *str, unsigned int length, int user_errors, int *pexit_code)
_gs_run_string_continue = gs.gsapi_run_string_continue
_gs_run_string_continue.argtypes = [c_void_p, c_char_p, c_uint, c_int, POINTER(c_int)]
_gs_run_string_continue.restype = c_int

# int gsapi_run_string_end (void *instance, int user_errors, int *pexit_code)
_gs_run_string_end = gs.gsapi_run_string_end
_gs_run_string_end.argtypes = [c_void_p, c_int, POINTER(c_int)]
_gs_run_string_end.restype = c_int

# int gsapi_run_string_with_length (void *instance, const char *str, unsigned int length, int user_errors, int *pexit_code)
_gs_run_string_with_length = gs.gsapi_run_string_with_length
_gs_run_string_with_length.argtypes = [c_void_p, c_char_p, c_uint, c_int, POINTER(c_int)]
_gs_run_string_with_length.restype = c_int

# int gsapi_run_string (void *instance, const char *str, int user_errors, int *pexit_code)
_gs_run_string = gs.gsapi_run_string
_gs_run_string.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_int)]
_gs_run_string.restype = c_int

# int gsapi_run_file (void *instance, const char *file_name, int user_errors, int *pexit_code)
_gs_run_file = gs.gsapi_run_file
_gs_run_file.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_int)]
_gs_run_file.restype = c_int

# Default callback handler for stdin/stdout/stderr
# You can override this by creating your own function and wrap it in GS_STDIOFUNC
def gs_output(handle, buffer, length):
    return length
# end gs_output

if sys.platform.startswith("win"):
    GS_STDIOFUNC = WINFUNCTYPE(c_int, c_void_p, c_char_p, c_int)
else:
    GS_STDIOFUNC = CFUNCTYPE(c_int, c_void_p, c_char_p, c_int)

log_func = GS_STDIOFUNC(gs_output)


# Set the callback functions for stdio The stdin callback function should return the number of characters read, 
# 0 for EOF, or -1 for error. The stdout and stderr callback functions should return the number of characters written. 

# int gsapi_set_stdio (void *instance, int(*stdin_fn)(void *caller_handle, char *buf, int len), 
#                   int(*stdout_fn)(void *caller_handle, const char *str, int len), 
#                   int(*stderr_fn)(void *caller_handle, const char *str, int len) ); 

# CMPFUNC = CFUNCTYPE(c_int, POINTER(c_int), POINTER(c_int))
_gs_set_stdio = gs.gsapi_set_stdio
#_gs_set_stdio.argtypes = [c_void_p, log_func, log_func, log_func]
_gs_set_stdio.restype = c_int

# END DLL DECLARATIONS

class GhostScriptError(Exception):
    def __init__(self, value, return_code = 0):
        Exception.__init__(self, value)
        self._value = value
        self._return_code = return_code
    def __str__(self):
        return repr("Ghostscript Error %d: %s" % (self._return_code, self._value))
    
    def ReturnCode(self):
        return self._return_code
# end GhostScriptError

# PDF compatibility levels - PDF/X-3, PDF/A is being worked on
# The default assumed for the GhostScript class is 1.4, which is
# also the default of the Ghostscript library
PDF_COMPAT_LEVELS = ["1.2", "1.3", "1.4"]

# Main Class for interfacing with Ghostscript
class GhostScript:
    def __init__(self, stdio_callback = None):
        # PDF Compatibility Level
        # Valid values are 1.2, 1.3, and 1.4
        self._pdfCompatLevel = "1.4"
        self._compressFonts = False
        self._instance_handle = c_void_p()
        
        self.__create_instance_handle()
        
        if stdio_callback is not None:
            _gs_set_stdio(self._instance_handle, stdio_callback, stdio_callback, stdio_callback)
    # end __init__
    
    def __create_instance_handle(self):
        # Create an instance handle to GS, which is used in
        # all other functions
        
        #if self._instance_handle != -1000:
        #    self.Shutdown()
        
        self._instance_handle = c_void_p()
        rc = c_int()
        rc = _gs_new_instance(byref(self._instance_handle), None)
        
        if rc > 0:
            raise GhostScriptError("Could not create instance handle to Ghostscript", rc)
    

    def SetPDFLevel(self, newLevel = "1.4"):
        # Set PDF compatibility level
        if newLevel not in PDF_COMPAT_LEVELS:
            newLevel = "1.4"
        
        self._pdfCompatLevel = newLevel
    # end SetPDFLevel
    
    def SetFontCompression(self, compressFonts = True):
        self._compressFonts = compressFonts
    # end SetFontCompression
    
    # Performs a clean shutdown of Ghostscript
    def Shutdown(self):
        try:
            rc = c_int()
            rc = _gs_exit(self._instance_handle)
        except:
            raise GhostScriptError("Bad exit from Ghostscript", rc)

        _gs_delete_instance(self._instance_handle)
    # end Shutdown
    
    def pdf2ps(self, input_file, output_file):
        # Converts a PDF to PostScript
        arguments = ["pdf2ps", "-dNOPAUSE", "-dBATCH", "-dSAFER", "-sDEVICE#pswrite", "-q", "-sOutputFile#%s" % output_file, input_file]
        gsargv, gsargc = self.__create_char_p_array(arguments)
        self.__run_cmd(gsargv, gsargc)
    # end pdf2ps
    
    def ps2pdf(self, input_file, output_file):
        # Converts a PostScript file into a PDF
        arguments = ["ps2pdf", "-q", "-dNOPAUSE", "-dBATCH", "-dSAFER", "-sDEVICE#pdfwrite", "-dCompatibilityLevel#%s" % self._pdfCompatLevel, "-sOutputFile#%s" % output_file]
        if self._compressFonts:
            arguments.append("-dCompressFonts=true")
        else:
            arguments.append("-dCompressFonts=false")
            
        arguments.append("-c")
        arguments.append(".setpdfwrite")
        arguments.append("-f")
        arguments.append(input_file)
        
        gsargv, gsargc = self.__create_char_p_array(arguments)
        self.__run_cmd(gsargv, gsargc)
    # end ps2pdf
    
    def __run_cmd(self, gsargv, gsargc):
    
        rc = c_int()
        rc = _gs_init(self._instance_handle, gsargc, byref(gsargv))
        if rc > 0:
            self.Shutdown()
            raise GhostScriptError("Unspecified Ghostscript Error", rc)
        elif rc <= -100:
            self.Shutdown()
            raise GhostScriptError("Fatal Error in Ghostscript", rc)

        self.Shutdown()
    # end __run_cmd
    
    def ps2epsi(self, input_file, output_file = ''):
        # ps2epsi currently does not work. Encapsulation code needs to be written
        # around the input file prior to working.

        arguments = ["ps2epsi", "-dNOPAUSE", "-q", "-dSAFER", "-dDELAYSAFER", input_file, output_file]
        gsargv, gsargc = self.__create_char_p_array(arguments)
        self.__run_cmd(gsargv, gsargc)
    # end ps2epsi
    
    def __create_char_p_array(self, string_list):
        StringArrayType = c_char_p * len(string_list)
        gsargv = StringArrayType(*string_list)
        gsargc = c_int(len(string_list))
        
        return (gsargv, gsargc)
    # end __create_char_p_array
    
# end class

if __name__ == '__main__':
    import glob
    
    ps_files = glob.glob(os.path.join("examples", "*.ps"))
    
    test_pdf = "test_pdf_input.pdf"
    
    for input_file in ps_files:
        output_file = os.path.splitext(input_file)[0] + ".pdf"
        
        print "Converting %s to %s" % (input_file, output_file)
        GS = GhostScript()        
        try:
            GS.ps2pdf(input_file, output_file)
        except GhostScriptError, gse:
            if gse.ReturnCode() <= -100:
                print gse
                print "Error processing file %s" % input_file
            else:
                print gse
            sys.exit()

    #print "Converting %s to test_pdf.ps..." % test_pdf
    #GS = GhostScript()
    #GS.pdf2ps(test_pdf, "test_pdf.ps")
    
    #GS = GhostScript()
    #print "Converting test_pdf.ps back to output2.pdf..."
    #GS.ps2pdf("test_pdf.ps", "output2.pdf")